from enum import Enum


class OutputFormat(Enum):
    MD = "md"
    STDOUT_MD = "stdout.md"
    READABLE_HTML = "html"
    RAW_HTML = "raw.html"

    def __str__(self):
        return self.value

    def is_file_output(self) -> bool:
        """Check if this format should be written to a file (vs stdout)."""
        return "stdout" not in self.value


class OutputFormatList:
    def __init__(self, formats: list[str]):
        self._formats: list[OutputFormat] = [OutputFormat(fmt) for fmt in formats]

    def __iter__(self):
        return iter(self._formats)

    def __contains__(self, item: OutputFormat) -> bool:
        return item in self._formats

    def __len__(self) -> int:
        return len(self._formats)

    def should_output_raw_html(self) -> bool:
        return OutputFormat.RAW_HTML in self._formats

    def should_output_readable_html(self) -> bool:
        return OutputFormat.READABLE_HTML in self._formats

    def should_output_markdown(self) -> bool:
        return OutputFormat.MD in self._formats or OutputFormat.STDOUT_MD in self._formats

    def should_output_markdown_file(self) -> bool:
        return OutputFormat.MD in self._formats

    def should_output_markdown_stdout(self) -> bool:
        return OutputFormat.STDOUT_MD in self._formats



# Tests that the outputs returned by Grabber.grab match exactly the requested output formats
from __future__ import annotations

import pytest
from clipit.clipper import Clipper
from clipit.core.output_format import OutputFormat


@pytest.mark.parametrize(
    "requested_formats, expected_keys",
    [
        ([(OutputFormat.MD.value)], {OutputFormat.MD}),
        ([(OutputFormat.STDOUT_MD.value)], {OutputFormat.STDOUT_MD}),
        (
            [OutputFormat.MD.value, OutputFormat.STDOUT_MD.value],
            {OutputFormat.MD, OutputFormat.STDOUT_MD},
        ),
    ],
)
def test_grab_outputs_match_requested_formats(requested_formats: list[str], expected_keys: set[OutputFormat]):
    url = "https://example.com/"
    grabber = Clipper()

    title, outputs = grabber.clip(
        url=url,
        use_readability_js=False,
        fallback_title="Untitled {date}",
        include_source=False,
        include_title=False,
        yaml_frontmatter=False,
        output_formats=requested_formats,
    )

    assert isinstance(title, str)

    returned_keys = set(outputs.keys())
    assert returned_keys == expected_keys, (
        f"Returned formats {sorted(map(str, returned_keys))} do not match requested {sorted(requested_formats)}"
    )

    # Additionally, ensure each requested format has non-empty content
    for fmt in expected_keys:
        assert fmt in outputs
        assert isinstance(outputs[fmt], str)
        assert outputs[fmt] != ""



from clipit.core.dtos import OutputFlags, RenderFlags
from clipit.core.misc import ClipitError
from clipit.core.output_format import OutputFormat, OutputFormatList

__all__ = ["OutputFormat", "OutputFormatList", "RenderFlags", "OutputFlags", "ClipitError"]



from readabilipy import simple_json_from_html_string

from clipit.core import ClipitError


def extract_readable_content_and_title(html_content, use_readability_js):
    try:
        rpy = simple_json_from_html_string(html_content, use_readability=use_readability_js)
        content_html = rpy.get("content") or ""

        # If readability.js fails, try again without it
        if not content_html and use_readability_js:
            rpy = simple_json_from_html_string(html_content, use_readability=False)
            content_html = rpy.get("content", "")
            if not content_html:
                raise ClipitError("No content found")

        content_html = content_html.replace(
            'href="about:blank/', 'href="../'
        )  # Fix for readability replacing ".." with "about:blank"
        title = (rpy.get("title") or "").strip()
    except Exception as e:
        raise ClipitError(f"Error processing HTML content: {e}")
    return (
        content_html,
        title,
    )



import re
from pathlib import Path
from urllib.parse import urlparse

import click

from clipit.core import ClipitError, OutputFlags, OutputFormat


def should_output_file(output_formats: dict[OutputFormat, str]) -> bool:
    return any(fmt.is_file_output() for fmt in output_formats)


def output(title: str, outputs: dict[OutputFormat, str], url: str, create_domain_subdir: bool, overwrite: bool):
    output_dir = None
    safe_title = None

    output_flags = OutputFlags(
        create_domain_subdir=create_domain_subdir,
        overwrite=overwrite,
    )

    if should_output_file(outputs):
        if output_flags.create_domain_subdir:
            output_dir = create_output_dir(url)
        else:
            output_dir = Path(".")
        safe_title = sanitize_filename(title)

    for format, output in outputs.items():
        if format.is_file_output():
            # output_dir and safe_title are only defined if we're saving to a file
            if output is not None and output_dir is not None and safe_title is not None:
                write_to_file(output, str(output_dir), safe_title, format.value, output_flags.overwrite)
        else:
            if output is not None:
                click.echo(output)


def sanitize_filename(filename):
    # Remove Obsidian-specific characters
    sanitized = re.sub(r"[#|\^\[\]]", "", filename)

    # Most conservative approach - remove all problematic characters including control characters
    sanitized = re.sub(r'[<>:"/\\|?*\x00-\x1F]', "", sanitized)

    # Handle Windows reserved filenames
    sanitized = re.sub(r"^(con|prn|aux|nul|com[0-9]|lpt[0-9])(\..*)?$", r"_\1\2", sanitized, flags=re.IGNORECASE)

    # Remove trailing spaces and periods
    sanitized = re.sub(r"[\s.]+$", "", sanitized)

    # Remove leading periods
    sanitized = re.sub(r"^\.+", "", sanitized)

    # Trim to 240 characters to leave room for extensions
    sanitized = sanitized[:240]

    return sanitized


def write_to_file(
    markdown_content: str,
    output_dir: str,
    safe_title: str,
    extension: str,
    overwrite: bool,
):
    output_file = Path(output_dir) / f"{safe_title}.{extension}"

    if not overwrite and output_file.exists():
        click.echo(f"File {output_file} already exists. Use --overwrite to replace it.")
        return

    try:
        with open(output_file, "w", encoding="utf-8") as f:
            f.write(markdown_content)
        click.echo(f"Saved {extension} content to {output_file}")
    except Exception as e:
        raise ClipitError(f"Error writing to file {output_file}: {e}")


def create_output_dir(url):
    parsed_url = urlparse(url)
    domain = parsed_url.netloc.replace("www.", "")
    if not domain:
        domain = "unknown_domain"
    output_dir = Path(".") / domain
    output_dir.mkdir(exist_ok=True, parents=True)

    return output_dir



import pytest
from clipit.core.writer import sanitize_filename


@pytest.mark.parametrize(
    "input_filename, expected_output",
    [
        ("invalid|file:name.txt", "invalidfilename.txt"),
        ("another/invalid\\name.txt", "anotherinvalidname.txt"),
        ("valid_name.txt", "valid_name.txt"),
    ],
)
def test_sanitize_filename_should_work(input_filename, expected_output):
    assert sanitize_filename(input_filename) == expected_output


def test_sanitize_filename_should_not_create_hidden_files():
    assert sanitize_filename(".NET Core") == "NET Core"



from dataclasses import dataclass


@dataclass
class RenderFlags:
    include_source: bool
    include_title: bool
    yaml_frontmatter: bool


@dataclass
class OutputFlags:
    create_domain_subdir: bool
    overwrite: bool



from clipit.core import OutputFormat, OutputFormatList
from clipit.core.dtos import RenderFlags
from clipit.core.writer import output
from clipit.grabbers import BaseGrabber, RedditGrabber

grabbers: list[BaseGrabber] = [RedditGrabber(), BaseGrabber()]


class Clipper:
    def __init__(self, user_agent: str | None = None):
        self.user_agent = user_agent

    def clip(
        self,
        url: str,
        use_readability_js: bool,
        fallback_title: str,
        include_source: bool,
        include_title: bool,
        yaml_frontmatter: bool,
        output_formats: list[str],
    ) -> tuple[str, dict[OutputFormat, str]]:
        grabber = next((g for g in grabbers if g.can_handle(url)), None)
        if grabber is None:
            raise ValueError("No grabber found for the given URL.")

        output_format_list: OutputFormatList = OutputFormatList(output_formats)
        render_flags = RenderFlags(
            include_source=include_source,
            include_title=include_title,
            yaml_frontmatter=yaml_frontmatter,
        )

        return grabber.grab(url, self.user_agent, use_readability_js, fallback_title, render_flags, output_format_list)

    def clip_and_save(
        self,
        url: str,
        use_readability_js: bool,
        fallback_title: str,
        include_source: bool,
        include_title: bool,
        yaml_frontmatter: bool,
        output_formats: list[str],
        create_domain_subdir: bool,
        overwrite: bool,
    ) -> None:
        title, outputs = self.clip(
            url, use_readability_js, fallback_title, include_source, include_title, yaml_frontmatter, output_formats
        )
        output(title, outputs, url, create_domain_subdir, overwrite)



import click

from clipit import ClipitError, Clipper, OutputFormat, __version__


@click.command()
@click.argument("url")
@click.option(
    "--user-agent",
    default=f"Clipit/{__version__}",
    help="The user agent reported when retrieving web pages",
    show_default=True,
)
@click.version_option(
    version=__version__,
    prog_name="Clipit",
    message="%(prog)s v%(version)s © 2025 Vlad Iliescu\n%(prog)s is licensed under the LGPL v3 License (https://www.gnu.org/licenses/lgpl-3.0.html)",
)
@click.option(
    "--yaml-frontmatter/--no-yaml-frontmatter",
    default=True,
    help="Include YAML front matter with metadata.",
    show_default=True,
)
@click.option(
    "--include-title/--no-include-title",
    default=True,
    help="Include the page title as an H1 heading.",
    show_default=True,
)
@click.option(
    "--include-source/--no-include-source",
    default=False,
    help="Include the page source.",
    show_default=True,
)
@click.option(
    "--fallback-title",
    default="Untitled {date}",
    help="Fallback title if no title is found. Use {date} for current date.",
    show_default=True,
)
@click.option(
    "--use-readability-js/--no-use-readability-js",
    default=True,
    help="Use Readability.js for processing pages, requires Node to be installed (recommended).",
    show_default=True,
)
@click.option(
    "--create-domain-subdir/--no-create-domain-subdir",
    default=True,
    help="Save the resulting file(s) in a subdirectory named after the domain.",
    show_default=True,
)
@click.option(
    "--overwrite/--no-overwrite",
    default=False,
    help="Overwrite existing files if they already exist.",
    show_default=True,
)
@click.option(
    "-f",
    "--format",
    "output_formats",
    multiple=True,
    default=[OutputFormat.MD.value],
    type=click.Choice([fmt.value for fmt in OutputFormat], case_sensitive=False),
    help="Which output format(s) to use when saving the content. Can be specified multiple times i.e. -f md -f html",
    show_default=True,
)
def main(
    url: str,
    user_agent: str,
    use_readability_js: bool,
    yaml_frontmatter: bool,
    include_title: bool,
    include_source: bool,
    fallback_title: str,
    create_domain_subdir: bool,
    output_formats: list[str],
    overwrite: bool,
) -> None:
    """
    Download a URL, convert it to Markdown/HTML with specified options, and save it to a file.
    """
    try:
        grabber = Clipper(user_agent=user_agent)
        grabber.clip_and_save(
            url=url,
            use_readability_js=use_readability_js,
            fallback_title=fallback_title,
            include_source=include_source,
            include_title=include_title,
            yaml_frontmatter=yaml_frontmatter,
            output_formats=list(output_formats),
            create_domain_subdir=create_domain_subdir,
            overwrite=overwrite,
        )
    except ClipitError as e:
        raise click.ClickException(str(e))


if __name__ == "__main__":
    main()



import json
from urllib.parse import urlparse, urlunparse

from clipit.core import ClipitError, RenderFlags
from clipit.core.downloader import download_html_content
from clipit.core.output_format import OutputFormat, OutputFormatList
from clipit.grabbers.base_grabber import BaseGrabber


class RedditGrabber(BaseGrabber):
    def can_handle(self, url: str) -> bool:
        domain = urlparse(url).netloc.lower()
        return domain == "www.reddit.com" or domain == "old.reddit.com"

    def grab(
        self,
        url: str,
        user_agent: str | None,
        use_readability_js: bool,
        fallback_title: str,
        render_flags: RenderFlags,
        output_formats: OutputFormatList,
    ) -> tuple[str, dict[OutputFormat, str]]:
        if (
            output_formats.should_output_raw_html()
            or output_formats.should_output_readable_html()
            or not output_formats.should_output_markdown()
        ):
            raise ClipitError("Reddit posts can only be converted to Markdown.")

        outputs = {}

        json_url = self._convert_to_json_url(url)
        json_content = json.loads(download_html_content(json_url, user_agent))

        title = json_content[0]["data"]["children"][0]["data"].get("title", None)
        title = self.post_process_title(title, fallback_title)

        markdown_content = self._reddit_json_to_markdown(json_content)
        markdown_content = self.post_process_markdown(url, title, markdown_content, render_flags)

        outputs[OutputFormat.MD] = markdown_content
        outputs[OutputFormat.STDOUT_MD] = markdown_content

        return title, outputs

    def _convert_to_json_url(self, url):
        parsed_url = urlparse(url)

        path = parsed_url.path.rstrip("/")
        new_path = f"{path}.json"

        json_url = urlunparse(parsed_url._replace(path=new_path))
        return json_url

    def _reddit_json_to_markdown(self, reddit_post_json):
        def parse_comments(comments_data, depth=0):
            comments_md = ""
            # Sort comments by score, highest first
            sorted_comments = sorted(comments_data, key=lambda x: x["data"].get("score", 0), reverse=True)
            for comment in sorted_comments:
                comment_data = comment["data"]
                author = comment_data.get("author", "[deleted]")
                score = comment_data.get("score", 0)
                body = comment_data.get("body", "").replace("\n", "\n" + "    " * (depth + 1))
                indentation = "    " * depth

                comments_md += f"{indentation}- **{author}** [{score} score]:\n{indentation}    {body}\n\n"

                # Check if 'replies' is a dict (has replies), and recursively parse them
                if isinstance(comment_data.get("replies"), dict):
                    nested_comments = comment_data["replies"]["data"]["children"]
                    comments_md += parse_comments(nested_comments, depth + 1)

            return comments_md

        try:
            # Extract post information
            post_data = reddit_post_json[0]["data"]["children"][0]["data"]
            selftext = post_data.get("selftext", "").replace("\n", "\n> ")
            post_url = post_data.get("url", "")  # needed for link posts
            author = post_data.get("author", "[deleted]")
            score = post_data.get("score", 0)

            markdown = f"**{author}** [{score} score]:\n> {selftext if selftext else post_url}\n\n"

            # Extract comments
            comments_data = reddit_post_json[1]["data"]["children"]
            markdown += "## Comments\n\n"
            markdown += parse_comments(comments_data)

        except Exception as e:
            raise ClipitError(f"Error converting Reddit JSON to Markdown: {str(e)}")

        return markdown



from importlib.metadata import PackageNotFoundError, version

from clipit.clipper import Clipper
from clipit.core import ClipitError, OutputFormat

try:
    __version__ = version("clipit")
except PackageNotFoundError:
    # Fallback when running from source without installation
    __version__ = "dev"

__all__ = [
    "__version__",
    "Clipper",
    "OutputFormat",
    "ClipitError",
]



from datetime import datetime

import yaml
from markdownify import (
    ATX,
    UNDERSCORE,
    MarkdownConverter,
    abstract_inline_conversion,  # type: ignore
)
from mdformat import text as mdformat_text


def try_include_title(include_title, markdown_content, title):
    if include_title:
        markdown_content = f"# {title}\n\n{markdown_content}"
    return markdown_content


def try_include_source(include_source, markdown_content, url):
    if include_source:
        markdown_content = f"[Source]({url})\n\n{markdown_content}"
    return markdown_content


def try_add_yaml_frontmatter(yaml_frontmatter: bool, markdown_content, title, url):
    if not yaml_frontmatter:
        return markdown_content

    metadata = {
        "title": title,
        "source": url,
        "date": datetime.now().strftime("%Y-%m-%d %H:%M"),
    }

    yaml_metadata = yaml.dump(metadata, sort_keys=False)
    markdown_content = f"---\n{yaml_metadata}---\n\n{markdown_content}"
    return markdown_content


class ClipitMarkdownConverter(MarkdownConverter):
    def convert_em(self, el, text, parent_tags):
        return self.convert_i(el, text, parent_tags)

    def convert_i(self, el, text, parent_tags):
        """I like my bolds ** and my italics _."""
        return abstract_inline_conversion(lambda s: UNDERSCORE)(self, el, text, parent_tags)


def convert_to_markdown(content_html):
    converter = ClipitMarkdownConverter(heading_style=ATX, bullets="-")
    markdown_content = converter.convert(content_html)
    pretty_markdown_content = mdformat_text(markdown_content)
    return pretty_markdown_content



from clipit.core.markdown_converter import convert_to_markdown


def test_converts_bold_and_italics_with_different_markup():
    html_content = "<strong>Bold Text</strong> and <i>Italic Text</i>"
    markdown_content = convert_to_markdown(html_content)

    assert "**Bold Text**" in markdown_content, "Bold text conversion failed"
    assert "_Italic Text_" in markdown_content, "Italic text conversion failed"


def test_converts_title_with_links():
    html_content = '<h2>TITLE (<a href="https://example.com">Link</a>)</h2>'
    markdown_content = convert_to_markdown(html_content)

    assert markdown_content == "## TITLE ([Link](https://example.com))\n", "Header conversion failed"



from datetime import datetime

from clipit.core import OutputFormat, OutputFormatList, RenderFlags
from clipit.core.downloader import download_html_content
from clipit.core.extractor import extract_readable_content_and_title
from clipit.core.markdown_converter import (
    convert_to_markdown,
    try_add_yaml_frontmatter,
    try_include_source,
    try_include_title,
)


class BaseGrabber:
    def can_handle(self, url: str) -> bool:
        return True

    def grab(
        self,
        url: str,
        user_agent: str | None,
        use_readability_js: bool,
        fallback_title: str,
        render_flags: RenderFlags,
        output_formats: OutputFormatList,
    ) -> tuple[str, dict[OutputFormat, str]]:
        outputs = {}

        html_content = download_html_content(url, user_agent)
        if output_formats.should_output_raw_html():
            outputs[OutputFormat.RAW_HTML] = html_content

        html_readable_content, title = extract_readable_content_and_title(html_content, use_readability_js)
        title = self.post_process_title(title, fallback_title)

        if output_formats.should_output_readable_html():
            outputs[OutputFormat.READABLE_HTML] = html_readable_content

        if output_formats.should_output_markdown():
            markdown_content = convert_to_markdown(html_readable_content)
            markdown_content = self.post_process_markdown(url, title, markdown_content, render_flags)

            if output_formats.should_output_markdown_file():
                outputs[OutputFormat.MD] = markdown_content

            if output_formats.should_output_markdown_stdout():
                outputs[OutputFormat.STDOUT_MD] = markdown_content

        return title, outputs

    def render_markdown(self, markdown_content):
        return markdown_content

    def handle_missing_title(self, title: str, fallback_title: str):
        if not title:
            title = fallback_title.format(date=datetime.now().strftime("%Y-%m-%d"))

        return title

    def post_process_markdown(
        self,
        url: str,
        title: str,
        markdown_content: str,
        render_flags: RenderFlags,
    ):
        markdown_content = try_include_source(render_flags.include_source, markdown_content, url)
        markdown_content = try_include_title(render_flags.include_title, markdown_content, title)
        markdown_content = try_add_yaml_frontmatter(render_flags.yaml_frontmatter, markdown_content, title, url)

        return markdown_content

    def post_process_title(self, title: str, fallback_title: str):
        title = self.handle_missing_title(title, fallback_title)

        return title



import shutil
import subprocess
from pathlib import Path


def test_cli_generates_md_and_html_and_cleans_up():
    project_root = Path(__file__).resolve().parents[1]
    # Using uvx console script; no direct script path required

    # Work in a temporary subdirectory inside the project so `uv` finds pyproject.toml
    workdir = project_root / ".tmp_cli_integration"
    if workdir.exists():
        shutil.rmtree(workdir)
    workdir.mkdir(parents=True, exist_ok=True)

    cmd = [
        "uvx",
        "--from",
        str(project_root),
        "clipit",
        "--no-create-domain-subdir",
        "-f",
        "md",
        "-f",
        "html",
        "https://example.com",
    ]

    expected_md = workdir / "Example Domain.md"
    expected_html = workdir / "Example Domain.html"

    try:
        proc = subprocess.run(
            cmd,
            cwd=workdir,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            timeout=120,
        )

        if proc.returncode != 0:
            raise AssertionError(f"CLI exited with {proc.returncode}.\nSTDOUT:\n{proc.stdout}\nSTDERR:\n{proc.stderr}")

        assert expected_md.exists(), "Expected Markdown file was not created"
        assert expected_html.exists(), "Expected HTML file was not created"
    finally:
        if workdir.exists():
            shutil.rmtree(workdir)



import requests
from requests import RequestException

from clipit.core import ClipitError


def download_html_content(url, user_agent: str | None) -> str:
    try:
        request_headers = {
            "User-Agent": user_agent,
            "Accept-Language": "en-US,en;q=0.9",
        }

        if user_agent is None:
            del request_headers["User-Agent"]

        response = requests.get(url, headers=request_headers)
        response.raise_for_status()
        html_content = response.content.decode("utf-8")
    except RequestException as e:
        raise ClipitError(f"Error downloading {url}: {e}")

    return html_content
